-- | default mime type table
--
-- this file is generated from file /etc/mime.types

module Text.XML.HXT.DOM.MimeTypeDefaults
where

-- | the table with the mapping from file name extensions to mime types

mimeTypeDefaults :: [(String, String)]
mimeTypeDefaults
  = [ ("123",   "application/vnd.lotus-1-2-3")
    , ("3ds",   "image/x-3ds")
    , ("3g2",   "video/x-3gpp2")
    , ("3gp",   "video/3gpp")
    , ("669",   "audio/x-mod")
    , ("BAY",   "image/x-dcraw")
    , ("BLEND", "application/x-blender")
    , ("BMQ",   "image/x-dcraw")
    , ("C",     "text/x-c++src")
    , ("CR2",   "image/x-dcraw")
    , ("CRW",   "image/x-dcraw")
    , ("CS1",   "image/x-dcraw")
    , ("CSSL",  "text/css")
    , ("DC2",   "image/x-dcraw")
    , ("DCR",   "image/x-dcraw")
    , ("FFF",   "image/x-dcraw")
    , ("K25",   "image/x-dcraw")
    , ("KDC",   "image/x-dcraw")
    , ("MOS",   "image/x-dcraw")
    , ("MRW",   "image/x-dcraw")
    , ("NEF",   "image/x-dcraw")
    , ("NSV",   "video/x-nsv")
    , ("ORF",   "image/x-dcraw")
    , ("PAR2",  "application/x-par2")
    , ("PEF",   "image/x-dcraw")
    , ("RAF",   "image/x-dcraw")
    , ("RDC",   "image/x-dcraw")
    , ("SRF",   "image/x-dcraw")
    , ("TTC",   "application/x-font-ttf")
    , ("X3F",   "image/x-dcraw")
    , ("XM",    "audio/x-mod")
    , ("Z",     "application/x-compress")
    , ("a",     "application/x-archive")
    , ("aac",   "audio/x-aac")
    , ("abw",   "application/x-abiword")
    , ("abw.CRASHED",   "application/x-abiword")
    , ("abw.gz",        "application/x-abiword")
    , ("ac3",   "audio/ac3")
    , ("adb",   "text/x-adasrc")
    , ("ads",   "text/x-adasrc")
    , ("afm",   "application/x-font-afm")
    , ("ag",    "image/x-applix-graphics")
    , ("ai",    "application/illustrator")
    , ("aif",   "audio/x-aiff")
    , ("aif",   "audio/x-aiff")
    , ("aifc",  "audio/x-aiff")
    , ("aiff",  "audio/x-aiff")
    , ("aiff",  "audio/x-aiff")
    , ("al",    "application/x-perl")
    , ("anim[1-9j]",    "video/x-anim")
    , ("aop",   "application/x-frontline")
    , ("arj",   "application/x-arj")
    , ("as",    "application/x-applix-spreadsheet")
    , ("asax",  "application/x-asax")
    , ("asc",   "text/plain")
    , ("ascx",  "application/x-ascx")
    , ("asf",   "video/x-ms-asf")
    , ("ashx",  "application/x-ashx")
    , ("asix",  "application/x-asix")
    , ("asmx",  "application/x-asmx")
    , ("asp",   "application/x-asp")
    , ("aspx",  "application/x-aspx")
    , ("asx",   "video/x-ms-asf")
    , ("au",    "audio/basic")
    , ("avi",   "video/x-msvideo")
    , ("aw",    "application/x-applix-word")
    , ("axd",   "application/x-axd")
    , ("bak",   "application/x-trash")
    , ("bay",   "image/x-dcraw")
    , ("bcpio", "application/x-bcpio")
    , ("bdf",   "application/x-font-bdf")
    , ("bib",   "text/x-bibtex")
    , ("bin",   "application/octet-stream")
    , ("bin",   "application/x-stuffit")
    , ("blend", "application/x-blender")
    , ("blender",       "application/x-blender")
    , ("bmp",   "image/bmp")
    , ("bmq",   "image/x-dcraw")
    , ("boo",   "text/x-boo")
    , ("bz",    "application/x-bzip")
    , ("bz",    "application/x-bzip")
    , ("bz2",   "application/x-bzip")
    , ("bz2",   "application/x-bzip")
    , ("c",     "text/x-csrc")
    , ("c++",   "text/x-c++src")
    , ("caves", "application/x-gnome-stones")
    , ("cc",    "text/x-c++src")
    , ("cdf",   "application/x-netcdf")
    , ("cdr",   "application/vnd.corel-draw")
    , ("cer",   "application/x-x509-ca-cert")
    , ("cert",  "application/x-x509-ca-cert")
    , ("cgi",   "application/x-cgi")
    , ("cgm",   "image/cgm")
    , ("chm",   "application/x-chm")
    , ("chrt",  "application/x-kchart")
    , ("cht",   "application/chemtool")
    , ("class", "application/x-java")
    , ("cls",   "text/x-tex")
    , ("cmbx",  "application/x-cmbx")
    , ("config",        "application/x-config")
    , ("connection",    "application/x-gnome-db-connection")
    , ("cpio",  "application/x-cpio")
    , ("cpio.gz",       "application/x-cpio-compressed")
    , ("cpp",   "text/x-c++src")
    , ("cr2",   "image/x-dcraw")
    , ("crt",   "application/x-x509-ca-cert")
    , ("crw",   "image/x-dcraw")
    , ("cs",    "text/x-csharp")
    , ("cs1",   "image/x-dcraw")
    , ("csh",   "application/x-csh")
    , ("css",   "text/css")
    , ("csv",   "text/x-comma-separated-values")
    , ("cue",   "application/x-cue")
    , ("cur",   "image/x-win-bitmap")
    , ("cxx",   "text/x-c++src")
    , ("d",     "text/x-dsrc")
    , ("dat",   "video/mpeg")
    , ("database",      "application/x-gnome-db-database")
    , ("dbf",   "application/x-dbase")
    , ("dc",    "application/x-dc-rom")
    , ("dc2",   "image/x-dcraw")
    , ("dcl",   "text/x-dcl")
    , ("dcm",   "application/dicom")
    , ("dcr",   "image/x-dcraw")
    , ("deb",   "application/x-deb")
    , ("der",   "application/x-x509-ca-cert")
    , ("desktop",       "application/x-desktop")
    , ("devhelp",       "application/x-devhelp")
    , ("dia",   "application/x-dia-diagram")
    , ("dif",   "video/dv")
    , ("diff",  "text/x-patch")
    , ("disco", "application/x-disco")
    , ("display",       "application/x-gdesklets-display")
    , ("djv",   "image/vnd.djvu")
    , ("djvu",  "image/vnd.djvu")
    , ("doc",   "application/msword")
    , ("docbook",       "application/docbook+xml")
    , ("dsl",   "text/x-dsl")
    , ("dtd",   "text/x-dtd")
    , ("dv",    "video/dv")
    , ("dvi",   "application/x-dvi")
    , ("dwg",   "image/vnd.dwg")
    , ("dxf",   "image/vnd.dxf")
    , ("ear",   "application/x-java-archive")
    , ("egon",  "application/x-egon")
    , ("el",    "text/x-emacs-lisp")
    , ("eps",   "image/x-eps")
    , ("epsf",  "image/x-eps")
    , ("epsi",  "image/x-eps")
    , ("etheme",        "application/x-e-theme")
    , ("etx",   "text/x-setext")
    , ("exe",   "application/x-executable")
    , ("exe",   "application/x-ms-dos-executable")
    , ("ez",    "application/andrew-inset")
    , ("f",     "text/x-fortran")
    , ("fff",   "image/x-dcraw")
    , ("fig",   "image/x-xfig")
    , ("fits",  "image/x-fits")
    , ("flac",  "audio/x-flac")
    , ("flc",   "video/x-flic")
    , ("fli",   "video/x-flic")
    , ("flw",   "application/x-kivio")
    , ("fo",    "text/x-xslfo")
    , ("g3",    "image/fax-g3")
    , ("gb",    "application/x-gameboy-rom")
    , ("gcrd",  "text/directory")
    , ("gen",   "application/x-genesis-rom")
    , ("gf",    "application/x-tex-gf")
    , ("gg",    "application/x-sms-rom")
    , ("gif",   "image/gif")
    , ("glabels",       "application/x-glabels")
    , ("glade", "application/x-glade")
    , ("gmo",   "application/x-gettext-translation")
    , ("gnc",   "application/x-gnucash")
    , ("gnucash",       "application/x-gnucash")
    , ("gnumeric",      "application/x-gnumeric")
    , ("gpg",   "application/pgp-encrypted")
    , ("gra",   "application/x-graphite")
    , ("gsf",   "application/x-font-type1")
    , ("gtar",  "application/x-gtar")
    , ("gz",    "application/x-gzip")
    , ("h",     "text/x-chdr")
    , ("h++",   "text/x-chdr")
    , ("hdf",   "application/x-hdf")
    , ("hh",    "text/x-c++hdr")
    , ("hp",    "text/x-chdr")
    , ("hpgl",  "application/vnd.hp-hpgl")
    , ("hs",    "text/x-haskell")
    , ("htm",   "text/html")
    , ("html",  "text/html")
    , ("ica",   "application/x-ica")
    , ("icb",   "image/x-icb")
    , ("ico",   "image/x-ico")
    , ("ics",   "text/calendar")
    , ("idl",   "text/x-idl")
    , ("ief",   "image/ief")
    , ("iff",   "image/x-iff")
    , ("il",    "text/x-msil")
    , ("ilbm",  "image/x-ilbm")
    , ("iso",   "application/x-cd-image")
    , ("it",    "audio/x-it")
    , ("jam",   "application/x-jamin")
    , ("jar",   "application/x-jar")
    , ("jar",   "application/x-java-archive")
    , ("java",  "text/x-java")
    , ("jng",   "image/x-jng")
    , ("jnlp",  "application/x-java-jnlp-file")
    , ("jp2",   "image/jpeg2000")
    , ("jpe",   "image/jpeg")
    , ("jpeg",  "image/jpeg")
    , ("jpg",   "image/jpeg")
    , ("jpr",   "application/x-jbuilder-project")
    , ("jpx",   "application/x-jbuilder-project")
    , ("js",    "application/x-javascript")
    , ("js",    "text/x-js")
    , ("k",     "application/x-tex-pk")
    , ("k25",   "image/x-dcraw")
    , ("karbon",        "application/x-karbon")
    , ("kdc",   "image/x-dcraw")
    , ("kdelnk",        "application/x-desktop")
    , ("kfo",   "application/x-kformula")
    , ("kil",   "application/x-killustrator")
    , ("kino",  "application/x-smil")
    , ("kon",   "application/x-kontour")
    , ("kpm",   "application/x-kpovmodeler")
    , ("kpr",   "application/x-kpresenter")
    , ("kpt",   "application/x-kpresenter")
    , ("kra",   "application/x-krita")
    , ("ksp",   "application/x-kspread")
    , ("kud",   "application/x-kugar")
    , ("kwd",   "application/x-kword")
    , ("kwt",   "application/x-kword")
    , ("la",    "application/x-shared-library-la")
    , ("lha",   "application/x-lha")
    , ("lhs",   "text/x-literate-haskell")
    , ("lhz",   "application/x-lhz")
    , ("log",   "text/x-log")
    , ("ltx",   "text/x-tex")
    , ("lwo",   "image/x-lwo")
    , ("lwob",  "image/x-lwo")
    , ("lws",   "image/x-lws")
    , ("lyx",   "application/x-lyx")
    , ("lzh",   "application/x-lha")
    , ("lzh",   "application/x-lha")
    , ("lzo",   "application/x-lzop")
    , ("m",     "text/x-objcsrc")
    , ("m15",   "audio/x-mod")
    , ("m3u",   "audio/x-mpegurl")
    , ("m4a",   "audio/x-m4a")
    , ("man",   "application/x-troff-man")
    , ("master",        "application/x-master-page")
    , ("md",    "application/x-genesis-rom")
    , ("mdp",   "application/x-mdp")
    , ("mds",   "application/x-mds")
    , ("mdsx",  "application/x-mdsx")
    , ("me",    "text/x-troff-me")
    , ("mergeant",      "application/x-mergeant")
    , ("mgp",   "application/x-magicpoint")
    , ("mid",   "audio/midi")
    , ("midi",  "audio/midi")
    , ("mif",   "application/x-mif")
    , ("mkv",   "application/x-matroska")
    , ("mm",    "text/x-troff-mm")
    , ("mml",   "text/mathml")
    , ("mng",   "video/x-mng")
    , ("moc",   "text/x-moc")
    , ("mod",   "audio/x-mod")
    , ("moov",  "video/quicktime")
    , ("mos",   "image/x-dcraw")
    , ("mov",   "video/quicktime")
    , ("movie", "video/x-sgi-movie")
    , ("mp2",   "video/mpeg")
    , ("mp3",   "audio/mpeg")
    , ("mpe",   "video/mpeg")
    , ("mpeg",  "video/mpeg")
    , ("mpg",   "video/mpeg")
    , ("mps",   "application/x-mps")
    , ("mrproject",     "application/x-planner")
    , ("mrw",   "image/x-dcraw")
    , ("ms",    "text/x-troff-ms")
    , ("msod",  "image/x-msod")
    , ("msx",   "application/x-msx-rom")
    , ("mtm",   "audio/x-mod")
    , ("n",     "text/x-nemerle")
    , ("n64",   "application/x-n64-rom")
    , ("nb",    "application/mathematica")
    , ("nc",    "application/x-netcdf")
    , ("nef",   "image/x-dcraw")
    , ("nes",   "application/x-nes-rom")
    , ("nsv",   "video/x-nsv")
    , ("o",     "application/x-object")
    , ("obj",   "application/x-tgif")
    , ("oda",   "application/oda")
    , ("odb",   "application/vnd.oasis.opendocument.database")
    , ("odc",   "application/vnd.oasis.opendocument.chart")
    , ("odf",   "application/vnd.oasis.opendocument.formula")
    , ("odg",   "application/vnd.oasis.opendocument.graphics")
    , ("odi",   "application/vnd.oasis.opendocument.image")
    , ("odm",   "application/vnd.oasis.opendocument.text-master")
    , ("odp",   "application/vnd.oasis.opendocument.presentation")
    , ("ods",   "application/vnd.oasis.opendocument.spreadsheet")
    , ("odt",   "application/vnd.oasis.opendocument.text")
    , ("ogg",   "application/ogg")
    , ("old",   "application/x-trash")
    , ("oleo",  "application/x-oleo")
    , ("orf",   "image/x-dcraw")
    , ("otg",   "application/vnd.oasis.opendocument.graphics-template")
    , ("oth",   "application/vnd.oasis.opendocument.text-web")
    , ("otp",   "application/vnd.oasis.opendocument.presentation-template")
    , ("ots",   "application/vnd.oasis.opendocument.spreadsheet-template")
    , ("ott",   "application/vnd.oasis.opendocument.text-template")
    , ("p",     "text/x-pascal")
    , ("p12",   "application/x-pkcs12")
    , ("p7s",   "application/pkcs7-signature")
    , ("par2",  "application/x-par2")
    , ("pas",   "text/x-pascal")
    , ("patch", "text/x-patch")
    , ("pbm",   "image/x-portable-bitmap")
    , ("pcd",   "image/x-photo-cd")
    , ("pcf",   "application/x-font-pcf")
    , ("pcf.Z", "application/x-font-type1")
    , ("pcf.gz",        "application/x-font-pcf")
    , ("pcl",   "application/vnd.hp-pcl")
    , ("pdb",   "application/vnd.palm")
    , ("pdb",   "application/x-palm-database")
    , ("pdf",   "application/pdf")
    , ("pef",   "image/x-dcraw")
    , ("pem",   "application/x-x509-ca-cert")
    , ("perl",  "application/x-perl")
    , ("pfa",   "application/x-font-type1")
    , ("pfb",   "application/x-font-type1")
    , ("pfx",   "application/x-pkcs12")
    , ("pgm",   "image/x-portable-graymap")
    , ("pgn",   "application/x-chess-pgn")
    , ("pgp",   "application/pgp")
    , ("pgp",   "application/pgp-encrypted")
    , ("php",   "application/x-php")
    , ("php3",  "application/x-php")
    , ("php4",  "application/x-php")
    , ("pict",  "image/x-pict")
    , ("pict1", "image/x-pict")
    , ("pict2", "image/x-pict")
    , ("pkr",   "application/pgp-keys")
    , ("pl",    "application/x-perl")
    , ("planner",       "application/x-planner")
    , ("pln",   "application/x-planperfect")
    , ("pls",   "audio/x-scpls")
    , ("pls",   "audio/x-scpls")
    , ("pm",    "application/x-perl")
    , ("png",   "image/png")
    , ("pnm",   "image/x-portable-anymap")
    , ("po",    "text/x-gettext-translation")
    , ("pot",   "application/vnd.ms-powerpoint")
    , ("pot",   "text/x-gettext-translation-template")
    , ("ppm",   "image/x-portable-pixmap")
    , ("pps",   "application/vnd.ms-powerpoint")
    , ("ppt",   "application/vnd.ms-powerpoint")
    , ("ppz",   "application/vnd.ms-powerpoint")
    , ("prc",   "application/x-palm-database")
    , ("prj",   "application/x-anjuta-project")
    , ("prjx",  "application/x-prjx")
    , ("ps",    "application/postscript")
    , ("ps.gz", "application/x-gzpostscript")
    , ("psd",   "image/x-psd")
    , ("psf",   "application/x-font-linux-psf")
    , ("psid",  "audio/prs.sid")
    , ("pto",   "application/x-ptoptimizer-script")
    , ("pw",    "application/x-pw")
    , ("py",    "text/x-python")
    , ("pyc",   "application/x-python-bytecode")
    , ("pyo",   "application/x-python-bytecode")
    , ("qif",   "application/x-qw")
    , ("qt",    "video/quicktime")
    , ("qtvr",  "video/quicktime")
    , ("ra",    "audio/vnd.rn-realaudio")
    , ("ra",    "audio/x-pn-realaudio")
    , ("raf",   "image/x-dcraw")
    , ("ram",   "audio/x-pn-realaudio")
    , ("ram",   "audio/x-pn-realaudio")
    , ("rar",   "application/x-rar")
    , ("rar",   "application/x-rar-compressed")
    , ("ras",   "image/x-cmu-raster")
    , ("rdc",   "image/x-dcraw")
    , ("rdf",   "text/rdf")
    , ("rdp",   "application/x-rdp")
    , ("rej",   "application/x-reject")
    , ("rem",   "application/x-remoting")
    , ("resources",     "application/x-resources")
    , ("resx",  "application/x-resourcesx")
    , ("rgb",   "image/x-rgb")
    , ("rle",   "image/rle")
    , ("rm",    "application/vnd.rn-realmedia")
    , ("rm",    "audio/x-pn-realaudio")
    , ("rmm",   "audio/x-pn-realaudio")
    , ("rms",   "application/vnd.rn-realmedia-secure")
    , ("rmvb",  "application/vnd.rn-realmedia-vbr")
    , ("rng",   "text/x-rng")
    , ("roff",  "application/x-troff")
    , ("rpm",   "application/x-rpm")
    , ("rss",   "text/rss")
    , ("rt",    "text/vnd.rn-realtext")
    , ("rtf",   "application/rtf")
    , ("rtx",   "text/richtext")
    , ("rv",    "video/vnd.rn-realvideo")
    , ("s3m",   "audio/x-s3m")
    , ("sam",   "application/x-amipro")
    , ("sc",    "application/x-sc")
    , ("scd",   "application/x-scribus")
    , ("scd.gz",        "application/x-scribus")
    , ("scm",   "text/x-scheme")
    , ("sda",   "application/vnd.stardivision.draw")
    , ("sdc",   "application/vnd.stardivision.calc")
    , ("sdd",   "application/vnd.stardivision.impress")
    , ("sdp",   "application/sdp")
    , ("sdp",   "application/vnd.stardivision.impress")
    , ("sds",   "application/vnd.stardivision.chart")
    , ("sdw",   "application/vnd.stardivision.writer")
    , ("sgi",   "image/x-sgi")
    , ("sgl",   "application/vnd.stardivision.writer")
    , ("sgm",   "text/sgml")
    , ("sgml",  "text/sgml")
    , ("sh",    "application/x-shellscript")
    , ("shar",  "application/x-shar")
    , ("siag",  "application/x-siag")
    , ("sid",   "audio/prs.sid")
    , ("sig",   "application/pgp-signature")
    , ("sik",   "application/x-trash")
    , ("sit",   "application/stuffit")
    , ("sit",   "application/x-stuffit")
    , ("skr",   "application/pgp-keys")
    , ("sla",   "application/x-scribus")
    , ("sla.gz",        "application/x-scribus")
    , ("slk",   "text/spreadsheet")
    , ("smd",   "application/vnd.stardivision.mail")
    , ("smf",   "application/vnd.stardivision.math")
    , ("smi",   "application/smil")
    , ("smi",   "application/x-smil")
    , ("smil",  "application/smil")
    , ("smil",  "application/x-smil")
    , ("sml",   "application/smil")
    , ("sms",   "application/x-sms-rom")
    , ("snd",   "audio/basic")
    , ("so",    "application/x-sharedlib")
    , ("soap",  "application/x-soap-remoting")
    , ("spd",   "application/x-font-speedo")
    , ("sql",   "text/x-sql")
    , ("src",   "application/x-wais-source")
    , ("srf",   "image/x-dcraw")
    , ("ssm",   "application/x-streamingmedia")
    , ("stc",   "application/vnd.sun.xml.calc.template")
    , ("std",   "application/vnd.sun.xml.draw.template")
    , ("sti",   "application/vnd.sun.xml.impress.template")
    , ("stm",   "audio/x-stm")
    , ("stw",   "application/vnd.sun.xml.writer.template")
    , ("sty",   "text/x-tex")
    , ("sun",   "image/x-sun-raster")
    , ("sv4cpio",       "application/x-sv4cpio")
    , ("sv4crc",        "application/x-sv4crc")
    , ("svg",   "image/svg+xml")
    , ("swf",   "application/x-shockwave-flash")
    , ("sxc",   "application/vnd.sun.xml.calc")
    , ("sxd",   "application/vnd.sun.xml.draw")
    , ("sxg",   "application/vnd.sun.xml.writer.global")
    , ("sxi",   "application/vnd.sun.xml.impress")
    , ("sxm",   "application/vnd.sun.xml.math")
    , ("sxw",   "application/vnd.sun.xml.writer")
    , ("sylk",  "text/spreadsheet")
    , ("t",     "application/x-troff")
    , ("tar",   "application/x-tar")
    , ("tar.Z", "application/x-compressed-tar")
    , ("tar.Z", "application/x-tarz")
    , ("tar.bz",        "application/x-bzip-compressed-tar")
    , ("tar.bz",        "application/x-bzip-compressed-tar")
    , ("tar.bz2",       "application/x-bzip-compressed-tar")
    , ("tar.bz2",       "application/x-bzip-compressed-tar")
    , ("tar.gz",        "application/x-compressed-tar")
    , ("tar.gz",        "application/x-compressed-tar")
    , ("tar.lzo",       "application/x-lzop-compressed-tar")
    , ("tar.lzo",       "application/x-tzo")
    , ("taz",   "application/x-compressed-tar")
    , ("tbz",   "application/x-bzip-compressed-tar")
    , ("tbz2",  "application/x-bzip-compressed-tar")
    , ("tcl",   "text/x-tcl")
    , ("tex",   "text/x-tex")
    , ("texi",  "text/x-texinfo")
    , ("texinfo",       "text/x-texinfo")
    , ("tga",   "image/x-tga")
    , ("tgz",   "application/x-compressed-tar")
    , ("tgz",   "application/x-compressed-tar")
    , ("theme", "application/x-theme")
    , ("tif",   "image/tiff")
    , ("tiff",  "image/tiff")
    , ("tk",    "text/x-tcl")
    , ("tm",    "text/x-texmacs")
    , ("toc",   "application/x-toc")
    , ("torrent",       "application/x-bittorrent")
    , ("tr",    "application/x-troff")
    , ("ts",    "application/x-linguist")
    , ("ts",    "text/x-texmacs")
    , ("tsv",   "text/tab-separated-values")
    , ("ttc",   "application/x-font-ttf")
    , ("ttf",   "application/x-font-ttf")
    , ("txt",   "text/plain")
    , ("tzo",   "application/x-lzop-compressed-tar")
    , ("tzo",   "application/x-tzo")
    , ("ui",    "application/x-designer")
    , ("uil",   "text/x-uil")
    , ("ult",   "audio/x-mod")
    , ("uni",   "audio/x-mod")
    , ("uri",   "text/x-uri")
    , ("url",   "text/x-uri")
    , ("ustar", "application/x-ustar")
    , ("vb",    "text/x-vb")
    , ("vcf",   "text/directory")
    , ("vcs",   "text/calendar")
    , ("vct",   "text/directory")
    , ("vob",   "video/mpeg")
    , ("voc",   "audio/x-voc")
    , ("vor",   "application/vnd.stardivision.writer")
    , ("war",   "application/x-java-archive")
    , ("wav",   "audio/x-wav")
    , ("wb1",   "application/x-quattro-pro")
    , ("wb1",   "application/x-quattropro")
    , ("wb2",   "application/x-quattro-pro")
    , ("wb2",   "application/x-quattropro")
    , ("wb3",   "application/x-quattro-pro")
    , ("wb3",   "application/x-quattropro")
    , ("wk1",   "application/vnd.lotus-1-2-3")
    , ("wk3",   "application/vnd.lotus-1-2-3")
    , ("wk4",   "application/vnd.lotus-1-2-3")
    , ("wks",   "application/vnd.lotus-1-2-3")
    , ("wmf",   "image/x-wmf")
    , ("wml",   "text/vnd.wap.wml")
    , ("wmv",   "video/x-ms-wmv")
    , ("wpd",   "application/vnd.wordperfect")
    , ("wpg",   "application/x-wpg")
    , ("wri",   "application/x-mswrite")
    , ("wrl",   "model/vrml")
    , ("wsdl",  "application/x-wsdl")
    , ("x3f",   "image/x-dcraw")
    , ("xac",   "application/x-gnucash")
    , ("xbel",  "application/x-xbel")
    , ("xbm",   "image/x-xbitmap")
    , ("xcf",   "image/x-xcf")
    , ("xcf.bz2",       "image/x-compressed-xcf")
    , ("xcf.gz",        "image/x-compressed-xcf")
    , ("xds",   "text/x-xds")
    , ("xhtml", "application/xhtml+xml")
    , ("xi",    "audio/x-xi")
    , ("xla",   "application/vnd.ms-excel")
    , ("xlc",   "application/vnd.ms-excel")
    , ("xld",   "application/vnd.ms-excel")
    , ("xll",   "application/vnd.ms-excel")
    , ("xlm",   "application/vnd.ms-excel")
    , ("xls",   "application/vnd.ms-excel")
    , ("xlt",   "application/vnd.ms-excel")
    , ("xlw",   "application/vnd.ms-excel")
    , ("xm",    "audio/x-xm")
    , ("xmi",   "text/x-xmi")
    , ("xml",   "text/xml")
    , ("xpl",   "audio/x-scpls")
    , ("xpm",   "image/x-xpixmap")
    , ("xsl",   "text/x-xsl")
    , ("xsl",   "text/x-xslt")
    , ("xslfo", "text/x-xslfo")
    , ("xslt",  "text/x-xslt")
    , ("xul",   "application/vnd.mozilla.xul+xml")
    , ("xwd",   "image/x-xwindowdump")
    , ("zabw",  "application/x-abiword")
    , ("zip",   "application/zip")
    , ("zoo",   "application/x-zoo")
   ]

